/* Modulo Lucas sequences.

   Copyright 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>

#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"

#include "mpm.h"
#include "mpm-impl.h"


/* change this to "#define TRACE(x) x" to get some traces */
#define TRACE(x)


/* Calculate the N'th terms in a Lucas sequence modulo M with parameters P,Q
   and D=P^2-4*Q.  M must be odd.

   The results calculated is placed into the Udest and Vdest parameters
   given, either of which can be NULL if not wanted.

   The value of P need not be between 0 and M-1.  Negative values are allowed.
   All values are understood modulo M for the purposes of calculation.

   The following formulae are used, taking three squares per bit.

        U[2k] = ((U[k]+V[k])^2 - U[k]^2 - V[k]^2) / 2
        V[2k] = (V[k]^2 + D * U[k]^2) / 2

        U[2k+1] = (P*U[2k] + V[2k]) / 2
        V[2k+1] = (D*U[2k] + P*V[2k]) / 2


   T2 is a second temporary used so that the mpz_mul's can have a separate
   destination, thereby avoiding an _mp_allocate_func call. */


#define TRACE_UV(str)


#define mpz_sub_si(w,x,y) \
  do {                    \
    if (y >= 0)           \
      mpz_sub_ui (w, x, y);     \
    else                        \
      mpz_add_ui (w, x, - (unsigned long) y);   \
  } while (0)



/*   void (*double_UV) _PROTO((struct __mplucm_struct *)); */
/*   void (*double_Vonly) _PROTO((struct __mplucm_struct *)); */
/*   void (*step_UV) _PROTO((struct __mplucm_struct *, int bit)); */

void
mpr_lucas_init (mpr_lucas_ptr l, mpm_srcptr m)
{
  mpr_init (l->U, m);
  mpr_init (l->V, m);
  mpr_init (l->T, m);
  mpr_init (l->T2, m);
  mpr_init (l->two, m);
  mpr_init (l->Pr, m);
  mpr_init (l->Qr, m);
  mpr_init (l->Dr, m);
}

void
mpr_lucas_clear (mpr_lucas_ptr l, mpm_srcptr m)
{
  mpr_clear (l->U, m);
  mpr_clear (l->V, m);
  mpr_clear (l->T, m);
  mpr_clear (l->T2, m);
  mpr_clear (l->two, m);
  mpr_clear (l->Pr, m);
  mpr_clear (l->Qr, m);
  mpr_clear (l->Dr, m);
}


void
mpr_lucas__q1_double_vonly (mpr_lucas_ptr l, mpm_srcptr m)
{
  /* V[2k] = V[k]^2 - 2 */
  mpr_mul (l->V, l->V, l->V, m);
  mpr_sub (l->V, l->V, l->two, m);
}


void
mpr_lucas__r_double (mpr_lucas_ptr l, mpm_srcptr m)
{
  mpr_mul (l->T, l->U, l->U, m);       /* U^2                  */

  mpr_mul (l->T2, l->V, l->V, m);      /* V^2                  */

  mpr_add (l->V, l->V, l->U, m);       /* (U+V)                */
  mpr_mul (l->U, l->V, l->V, m);       /* (U+V)^2              */

  mpr_sub (l->U, l->U, l->T, m);       /* (U+V)^2 - U^2        */
  mpr_sub (l->U, l->U, l->T2, m);      /* (U+V)^2 - U^2 - V^2  */

  mpr_mul (l->T, l->T, l->Dr, m);      /* D*U^2                */
  mpr_add (l->T2, l->T2, l->T, m);     /* V^2 + D*U^2          */

  mpr_div_2exp (l->U, l->U, 1L, m);    /* U[2k] = ((U+V)^2 - U^2 - V^2)/2 */

  mpr_div_2exp (l->T2, l->T2, 1L, m);  /* V[2k] = (V^2 + D*U^2)/2 */

  mpr_swap (l->T2, l->V, m);
}

void
mpr_lucas__r_step (mpr_lucas_ptr l, mpm_srcptr m)
{
  /* T = U[2k+1] = (P*U[2k] + V[2k]) / 2 */
  mpr_mul (l->T, l->U, l->Pr, m);
  mpr_add (l->T, l->T, l->V, m);
  mpr_div_2exp (l->T, l->T, 1L, m);

  /* V[2k+1] = (D*U[2k] + P*V[2k]) / 2 */
  mpr_mul (l->U, l->U, l->Dr, m);
  mpr_mul (l->V, l->V, l->Pr, m);
  mpr_add (l->V, l->V, l->U, m);
  mpr_div_2exp (l->V, l->V, 1L, m);

  mpr_swap (l->U, l->T, m);
}

void
mpr_lucas_pq (mpr_lucas_ptr l, mpz_srcptr p, mpz_srcptr q, mpm_srcptr m)
{
  mpz_t  d, t;

  mpr_set_z (l->Pr, p, m);
  mpr_set_z (l->Qr, q, m);

  /* D = P^2 + 4*Q */
  mpz_init (d);
  mpz_init (t);
  mpz_mul (d, p, p);
  mpz_mul_2exp (t, q, 2);
  mpz_add (d, d, t);
  l->jac_DN = mpz_kronecker (d, mpm_modref(m));
  mpr_set_z (l->Dr, d, m);
  mpz_clear (d);
  mpz_clear (t);

  l->_mp_double = mpr_lucas__r_double;
  l->_mp_double_vonly = mpr_lucas__r_double;
  l->_mp_step = mpr_lucas__r_step;
}


void
mpr_lucas__si_double (mpr_lucas_ptr l, mpm_srcptr m)
{
  mpr_mul (l->T, l->U, l->U, m);       /* U^2                  */

  mpr_mul (l->T2, l->V, l->V, m);      /* V^2                  */

  mpr_add (l->V, l->V, l->U, m);       /* (U+V)                */
  mpr_mul (l->U, l->V, l->V, m);       /* (U+V)^2              */

  mpr_sub (l->U, l->U, l->T, m);       /* (U+V)^2 - U^2        */
  mpr_sub (l->U, l->U, l->T2, m);      /* (U+V)^2 - U^2 - V^2  */

  mpr_mul_si (l->T, l->T, l->Dl, m);   /* D*U^2                */
  mpr_add (l->T2, l->T2, l->T, m);     /* V^2 + D*U^2          */

  mpr_div_2exp (l->U, l->U, 1L, m);    /* U[2k] = ((U+V)^2 - U^2 - V^2)/2 */

  mpr_div_2exp (l->T2, l->T2, 1L, m);  /* V[2k] = (V^2 + D*U^2)/2 */

  mpr_swap (l->T2, l->V, m);
}

void
mpr_lucas__si_step (mpr_lucas_ptr l, mpm_srcptr m)
{
  /* T = U[2k+1] = (P*U[2k] + V[2k]) / 2 */
  mpr_mul_si (l->T, l->U, l->Pl, m);
  mpr_add (l->T, l->T, l->V, m);
  mpr_div_2exp (l->T, l->T, 1L, m);

  /* V[2k+1] = (D*U[2k] + P*V[2k]) / 2 */
  mpr_mul_si (l->U, l->U, l->Dl, m);
  mpr_mul_si (l->V, l->V, l->Pl, m);
  mpr_add (l->V, l->V, l->U, m);
  mpr_div_2exp (l->V, l->V, 1L, m);

  mpr_swap (l->U, l->T, m);
}


void
mpr_lucas_pq_si (mpr_lucas_ptr l, long p, long q, mpm_srcptr m)
{
  l->Pl = p;
  l->Ql = q;
  l->Dl = p*p - 4*q;

  l->jac_DN = mpz_si_kronecker (l->Dl, mpm_modref(m));

  mpr_set_si (l->Pr, p, m);

  l->_mp_double = mpr_lucas__si_double;
  l->_mp_step = mpr_lucas__si_step;

  if (q == 1)
    {
      l->_mp_double_vonly = mpr_lucas__q1_double_vonly;
      mpr_set_ui (l->two, 2L, m);
    }
  else
    l->_mp_double_vonly = mpr_lucas__si_double;
}





#define mpr_lucas_double(l,m)   do { (*(l)->_mp_double) (l, m); } while (0)
#define mpr_lucas_double_vonly(l,m)   do { (*(l)->_mp_double_vonly) (l, m); } while (0)
#define mpr_lucas_step(l,m)     do { (*(l)->_mp_step) (l, m); } while (0)


void
mpr_lucas_pow (mpr_lucas_ptr l, mpz_srcptr e, mpm_srcptr m)
{
  mp_size_t  esize, i;
  mp_srcptr  ep;
  mp_limb_t  elimb, mask;
  int        lz;
  
  /*   TRACE (mpz_trace ("  P", P); */
  /*          mpz_trace ("  Q", Q); */
  /*          mpz_trace ("  D", D); */
  /*          mpz_trace ("  N", N); */
  /*          mpz_trace ("  M", M)); */

  esize = SIZ(e);
  if (esize < 0)
    abort();

  if (esize == 0)
    {
      /* U[0] = 0, V[0] = 2 (mod M) */
      mpr_set_ui (l->U, 0L, m);
      mpr_set_ui (l->V, 2L, m);
      return;
    }

  mpr_set_ui (l->U, 1L, m);   /* U[1] = 1 */
  mpr_set (l->V, l->Pr, m);   /* V[1] = P */
  TRACE_UV ("At k=1");

  ep = PTR(e);
  i = esize-1;
  elimb = ep[i];
  count_leading_zeros (lz, elimb);
  mask = CNST_LIMB(1) << (BITS_PER_MP_LIMB - 1 - lz);

  for (;;)
    {
      mask >>= 1;
      if (mask == 0)
        {
          i--;
          if (i < 0)
            break;
          elimb = ep[i];
          mask = GMP_NUMB_HIGHBIT;
        }

      mpr_lucas_double (l, m);
      TRACE_UV ("Double");

      if ((elimb & mask) != 0)
        {
          mpr_lucas_step (l, m);
          TRACE_UV ("Double");
        }
    }

  TRACE_UV("Final");
}


int
mpr_lprp_p (mpr_lucas_ptr l, mpm_srcptr m)
{
  mpz_t  e;
  int    ret;

  TRACE (printf ("mpr_lprp_p\n"));

  mpz_init (e);
  mpz_sub_si (e, mpm_modref(m), l->jac_DN);
  mpr_lucas_pow (l, e, m);
  ret = mpr_equal_one (l->U, m);
  mpz_clear (e);
  return ret;
}

int
mpr_elprp_p (mpr_lucas_ptr l, mpm_srcptr m)
{
  abort ();
  return 0;
}

int
mpr_slprp_p (mpr_lucas_ptr l, mpm_srcptr m)
{
  abort ();
  return 0;
}
