/* mpn_mulhighhalf_basecase -- base case high half multiply.

   Copyright 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"
#include "mpm.h"
#include "mpm-impl.h"


/* wp[2,1,0] = sum(xp[i]*yp[n-1-i]), ie. an inner product, but with the
   indices reversed.
   Not sure if it's best to use nails during the summation, or accumulate
   as full limbs and nailify at the end.  */
void
mpn_revinnprod_n (mp_ptr wp, mp_srcptr xp, mp_srcptr yp, mp_size_t size)
{
  mp_size_t  i;
  mp_limb_t  w2, w1, w0, hi, lo;

  ASSERT (size >= 1);

  yp += size-1;
  umul_ppmm (w1, w0, *xp, *yp << GMP_NAIL_BITS);
  w2 = 0;
  xp++;
  yp--;

  for (i = 1; i < size; i++)
    {
      umul_ppmm (hi, lo, *xp, *yp << GMP_NAIL_BITS);
      lo >>= GMP_NAIL_BITS;
      ADD_332 (w2,w1,w0, w2,w1,w0, hi,lo);
      xp++;
      yp--;
    }

  wp[0] = w0;
  wp[1] = w1;
  wp[2] = w2;
}


void
mpn_mulhighhalf_basecase (mp_ptr wp, mp_srcptr xp, mp_srcptr yp, mp_size_t size)
{
  ASSERT (size >= 1);
  ASSERT (! MPN_OVERLAP_P (wp, 2*size, xp, size));
  ASSERT (! MPN_OVERLAP_P (wp, 2*size, yp, size));

  if (size == 1)
    {
      mp_limb_t  dummy;
      umul_ppmm (wp[0], dummy, xp[0], yp[0]);
      return;
    }

  if (size == 2)
    {
      mp_limb_t  w3, w2, w1, w0, hi, lo;
      mp_limb_t  x0;
      mp_limb_t  y0;
      mp_limb_t  x1 = xp[1];
      mp_limb_t  y1 = yp[1];

      umul_ppmm (w3, w2, x1, y1 << GMP_NAIL_BITS);
      w2 >>= GMP_NAIL_BITS;

      x0 = xp[0];
      umul_ppmm (hi, w1, x0, y1 << GMP_NAIL_BITS);
      w1 >>= GMP_NAIL_BITS;
      ADD_221 (w3,w2, w3,w2, hi);

      y0 = yp[0];
      umul_ppmm (hi, lo, x1, y0 << GMP_NAIL_BITS);
      lo >>= GMP_NAIL_BITS;
      ADD_332 (w3,w2,w1, w3,w2,w1, hi,lo);

      if (UNLIKELY (w1 == GMP_NUMB_MAX))
        {
          umul_ppmm (hi, w0, x0, y0 << GMP_NAIL_BITS);
          ADD_331 (w3,w2,w1, w3,w2,w1, hi);
        }
      wp[2] = w2;
      wp[3] = w3;
      return;
    }

#if HAVE_NATIVE_mpn_mul_2
  mpn_mul_2 (wp, xp, size, yp);
  wp += 2, yp += 2, size -= 2;
#else
  mpn_mul_1 (wp, xp, size, yp[0]);
  wp += 1, yp += 1, size -= 1;
#endif

#if HAVE_NATIVE_mpn_addmul_2
  while (size >= 2)
    {
      mpn_addmul_2 (wp, xp, size, yp);
      wp += 2, xp += 2, size -= 2;
    }
  if (size != 0)
    wp[0] += xp[0] * yp[0];
#else
  while (size > 1)
    {
      mpn_addmul_1 (wp, xp, size, yp[0]);
      wp += 1, yp += 1, size--;
    }
  wp[0] += xp[0] * yp[0];
#endif
}
